package org.eclipse.jdt.internal.compiler.apt.model;

package .tmp;

import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.AnnotationValue;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.type.DeclaredType;
import javax.lang.model.type.MirroredTypeException;
import javax.lang.model.type.MirroredTypesException;
import javax.lang.model.type.TypeMirror;
import org.eclipse.jdt.internal.compiler.apt.dispatch.BaseProcessingEnvImpl;
import org.eclipse.jdt.internal.compiler.apt.model.AnnotationMemberValue;
import org.eclipse.jdt.internal.compiler.apt.model.ExecutableElementImpl;
import org.eclipse.jdt.internal.compiler.apt.model.Factory;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.ArrayBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.ElementValuePair;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;

public class AnnotationMirrorImpl implements AnnotationMirror, InvocationHandler {
  public final BaseProcessingEnvImpl _env;
  
  public final AnnotationBinding _binding;
  
  AnnotationMirrorImpl(BaseProcessingEnvImpl env, AnnotationBinding binding) {
    this._env = env;
    this._binding = binding;
  }
  
  public boolean equals(Object obj) {
    if (obj instanceof org.eclipse.jdt.internal.compiler.apt.model.AnnotationMirrorImpl) {
      if (this._binding == null)
        return (((org.eclipse.jdt.internal.compiler.apt.model.AnnotationMirrorImpl)obj)._binding == null); 
      return equals(this._binding, ((org.eclipse.jdt.internal.compiler.apt.model.AnnotationMirrorImpl)obj)._binding);
    } 
    return (obj == null) ? false : obj.equals(this);
  }
  
  private static boolean equals(AnnotationBinding annotationBinding, AnnotationBinding annotationBinding2) {
    if (annotationBinding.getAnnotationType() != annotationBinding2.getAnnotationType())
      return false; 
    ElementValuePair[] elementValuePairs = annotationBinding.getElementValuePairs();
    ElementValuePair[] elementValuePairs2 = annotationBinding2.getElementValuePairs();
    int length = elementValuePairs.length;
    if (length != elementValuePairs2.length)
      return false; 
    for (int i = 0; i < length; i++) {
      ElementValuePair pair = elementValuePairs[i];
      int j = 0;
      while (true) {
        if (j >= length)
          return false; 
        ElementValuePair pair2 = elementValuePairs2[j];
        if (pair.binding == pair2.binding) {
          if (pair.value == null) {
            if (pair2.value == null)
              break; 
            return false;
          } 
          if (pair2.value == null)
            return false; 
          if (pair2.value instanceof Object[] && pair.value instanceof Object[]) {
            if (!Arrays.equals((Object[])pair.value, (Object[])pair2.value))
              return false; 
            break;
          } 
          if (!pair2.value.equals(pair.value))
            return false; 
          break;
        } 
        j++;
      } 
    } 
    return true;
  }
  
  public DeclaredType getAnnotationType() {
    return (DeclaredType)this._env.getFactory().newTypeMirror((Binding)this._binding.getAnnotationType());
  }
  
  public Map<? extends ExecutableElement, ? extends AnnotationValue> getElementValues() {
    if (this._binding == null)
      return Collections.emptyMap(); 
    ElementValuePair[] pairs = this._binding.getElementValuePairs();
    Map<ExecutableElement, AnnotationValue> valueMap = 
      new LinkedHashMap<>(pairs.length);
    byte b;
    int i;
    ElementValuePair[] arrayOfElementValuePair1;
    for (i = (arrayOfElementValuePair1 = pairs).length, b = 0; b < i; ) {
      ElementValuePair pair = arrayOfElementValuePair1[b];
      MethodBinding method = pair.getMethodBinding();
      if (method != null) {
        ExecutableElementImpl executableElementImpl = new ExecutableElementImpl(this._env, method);
        AnnotationMemberValue annotationMemberValue = new AnnotationMemberValue(this._env, pair.getValue(), method);
        valueMap.put(executableElementImpl, annotationMemberValue);
      } 
      b++;
    } 
    return Collections.unmodifiableMap(valueMap);
  }
  
  public Map<? extends ExecutableElement, ? extends AnnotationValue> getElementValuesWithDefaults() {
    if (this._binding == null)
      return Collections.emptyMap(); 
    ElementValuePair[] pairs = this._binding.getElementValuePairs();
    ReferenceBinding annoType = this._binding.getAnnotationType();
    Map<ExecutableElement, AnnotationValue> valueMap = 
      new LinkedHashMap<>();
    byte b;
    int i;
    MethodBinding[] arrayOfMethodBinding;
    for (i = (arrayOfMethodBinding = annoType.methods()).length, b = 0; b < i; ) {
      MethodBinding method = arrayOfMethodBinding[b];
      boolean foundExplicitValue = false;
      for (int j = 0; j < pairs.length; j++) {
        MethodBinding explicitBinding = pairs[j].getMethodBinding();
        if (method == explicitBinding) {
          ExecutableElementImpl executableElementImpl = new ExecutableElementImpl(this._env, explicitBinding);
          AnnotationMemberValue annotationMemberValue = new AnnotationMemberValue(this._env, pairs[j].getValue(), explicitBinding);
          valueMap.put(executableElementImpl, annotationMemberValue);
          foundExplicitValue = true;
          break;
        } 
      } 
      if (!foundExplicitValue) {
        Object defaultVal = method.getDefaultValue();
        if (defaultVal != null) {
          ExecutableElementImpl executableElementImpl = new ExecutableElementImpl(this._env, method);
          AnnotationMemberValue annotationMemberValue = new AnnotationMemberValue(this._env, defaultVal, method);
          valueMap.put(executableElementImpl, annotationMemberValue);
        } 
      } 
      b++;
    } 
    return Collections.unmodifiableMap(valueMap);
  }
  
  public int hashCode() {
    if (this._binding == null)
      return this._env.hashCode(); 
    return this._binding.hashCode();
  }
  
  public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
    if (this._binding == null)
      return null; 
    String methodName = method.getName();
    if (args == null || args.length == 0) {
      if (methodName.equals("hashCode"))
        return Integer.valueOf(hashCode()); 
      if (methodName.equals("toString"))
        return toString(); 
      if (methodName.equals("annotationType"))
        return proxy.getClass().getInterfaces()[0]; 
    } else if (args.length == 1 && methodName.equals("equals")) {
      return Boolean.valueOf(equals(args[0]));
    } 
    if (args != null && args.length != 0)
      throw new NoSuchMethodException("method " + method.getName() + formatArgs(args) + " does not exist on annotation " + toString()); 
    MethodBinding methodBinding = getMethodBinding(methodName);
    if (methodBinding == null)
      throw new NoSuchMethodException("method " + method.getName() + "() does not exist on annotation" + toString()); 
    Object actualValue = null;
    boolean foundMethod = false;
    ElementValuePair[] pairs = this._binding.getElementValuePairs();
    byte b;
    int i;
    ElementValuePair[] arrayOfElementValuePair1;
    for (i = (arrayOfElementValuePair1 = pairs).length, b = 0; b < i; ) {
      ElementValuePair pair = arrayOfElementValuePair1[b];
      if (methodName.equals(new String(pair.getName()))) {
        actualValue = pair.getValue();
        foundMethod = true;
        break;
      } 
      b++;
    } 
    if (!foundMethod)
      actualValue = methodBinding.getDefaultValue(); 
    Class<?> expectedType = method.getReturnType();
    TypeBinding actualType = methodBinding.returnType;
    return getReflectionValue(actualValue, actualType, expectedType);
  }
  
  public String toString() {
    TypeMirror decl = getAnnotationType();
    StringBuilder sb = new StringBuilder();
    sb.append('@');
    sb.append(decl.toString());
    Map<? extends ExecutableElement, ? extends AnnotationValue> values = getElementValues();
    if (!values.isEmpty()) {
      sb.append('(');
      boolean first = true;
      for (Map.Entry<? extends ExecutableElement, ? extends AnnotationValue> e : values.entrySet()) {
        if (!first)
          sb.append(", "); 
        first = false;
        sb.append(((ExecutableElement)e.getKey()).getSimpleName());
        sb.append(" = ");
        sb.append(((AnnotationValue)e.getValue()).toString());
      } 
      sb.append(')');
    } 
    return sb.toString();
  }
  
  private String formatArgs(Object[] args) {
    StringBuilder builder = new StringBuilder(args.length * 8 + 2);
    builder.append('(');
    for (int i = 0; i < args.length; i++) {
      if (i > 0)
        builder.append(", "); 
      builder.append(args[i].getClass().getName());
    } 
    builder.append(')');
    return builder.toString();
  }
  
  private MethodBinding getMethodBinding(String name) {
    ReferenceBinding annoType = this._binding.getAnnotationType();
    MethodBinding[] methods = annoType.getMethods(name.toCharArray());
    byte b;
    int i;
    MethodBinding[] arrayOfMethodBinding1;
    for (i = (arrayOfMethodBinding1 = methods).length, b = 0; b < i; ) {
      MethodBinding method = arrayOfMethodBinding1[b];
      if (method.parameters.length == 0)
        return method; 
      b++;
    } 
    return null;
  }
  
  private Object getReflectionValue(Object actualValue, TypeBinding actualType, Class<?> expectedType) {
    if (expectedType == null)
      return null; 
    if (actualValue == null)
      return Factory.getMatchingDummyValue(expectedType); 
    if (expectedType.isArray()) {
      if (Class.class.equals(expectedType.getComponentType())) {
        if (actualType.isArrayType() && actualValue instanceof Object[] && 
          (((ArrayBinding)actualType).leafComponentType.erasure()).id == 16) {
          Object[] bindings = (Object[])actualValue;
          List<TypeMirror> mirrors = new ArrayList<>(bindings.length);
          for (int i = 0; i < bindings.length; i++) {
            if (bindings[i] instanceof TypeBinding)
              mirrors.add(this._env.getFactory().newTypeMirror((Binding)bindings[i])); 
          } 
          throw new MirroredTypesException(mirrors);
        } 
        return null;
      } 
      return convertJDTArrayToReflectionArray(actualValue, actualType, expectedType);
    } 
    if (Class.class.equals(expectedType)) {
      if (actualValue instanceof TypeBinding) {
        TypeMirror mirror = this._env.getFactory().newTypeMirror((Binding)actualValue);
        throw new MirroredTypeException(mirror);
      } 
      return null;
    } 
    return convertJDTValueToReflectionType(actualValue, actualType, expectedType);
  }
  
  private Object convertJDTArrayToReflectionArray(Object jdtValue, TypeBinding jdtType, Class<?> expectedType) {
    Object[] jdtArray;
    assert expectedType != null && expectedType.isArray();
    if (!jdtType.isArrayType())
      return null; 
    if (jdtValue != null && !(jdtValue instanceof Object[])) {
      jdtArray = (Object[])Array.newInstance(jdtValue.getClass(), 1);
      jdtArray[0] = jdtValue;
    } else {
      jdtArray = (Object[])jdtValue;
    } 
    TypeBinding jdtLeafType = jdtType.leafComponentType();
    Class<?> expectedLeafType = expectedType.getComponentType();
    int length = jdtArray.length;
    Object returnArray = Array.newInstance(expectedLeafType, length);
    for (int i = 0; i < length; i++) {
      Object jdtElementValue = jdtArray[i];
      if (expectedLeafType.isPrimitive() || String.class.equals(expectedLeafType)) {
        if (jdtElementValue instanceof Constant) {
          if (boolean.class.equals(expectedLeafType)) {
            Array.setBoolean(returnArray, i, ((Constant)jdtElementValue).booleanValue());
          } else if (byte.class.equals(expectedLeafType)) {
            Array.setByte(returnArray, i, ((Constant)jdtElementValue).byteValue());
          } else if (char.class.equals(expectedLeafType)) {
            Array.setChar(returnArray, i, ((Constant)jdtElementValue).charValue());
          } else if (double.class.equals(expectedLeafType)) {
            Array.setDouble(returnArray, i, ((Constant)jdtElementValue).doubleValue());
          } else if (float.class.equals(expectedLeafType)) {
            Array.setFloat(returnArray, i, ((Constant)jdtElementValue).floatValue());
          } else if (int.class.equals(expectedLeafType)) {
            Array.setInt(returnArray, i, ((Constant)jdtElementValue).intValue());
          } else if (long.class.equals(expectedLeafType)) {
            Array.setLong(returnArray, i, ((Constant)jdtElementValue).longValue());
          } else if (short.class.equals(expectedLeafType)) {
            Array.setShort(returnArray, i, ((Constant)jdtElementValue).shortValue());
          } else if (String.class.equals(expectedLeafType)) {
            Array.set(returnArray, i, ((Constant)jdtElementValue).stringValue());
          } 
        } else {
          Factory.setArrayMatchingDummyValue(returnArray, i, expectedLeafType);
        } 
      } else if (expectedLeafType.isEnum()) {
        Object returnVal = null;
        if (jdtLeafType != null && jdtLeafType.isEnum() && jdtElementValue instanceof FieldBinding) {
          FieldBinding binding = (FieldBinding)jdtElementValue;
          try {
            Field returnedField = null;
            returnedField = expectedLeafType.getField(new String(binding.name));
            if (returnedField != null)
              returnVal = returnedField.get(null); 
          } catch (NoSuchFieldException noSuchFieldException) {
          
          } catch (IllegalAccessException illegalAccessException) {}
        } 
        Array.set(returnArray, i, returnVal);
      } else if (expectedLeafType.isAnnotation()) {
        Object returnVal = null;
        if (jdtLeafType.isAnnotationType() && jdtElementValue instanceof AnnotationBinding) {
          org.eclipse.jdt.internal.compiler.apt.model.AnnotationMirrorImpl annoMirror = 
            (org.eclipse.jdt.internal.compiler.apt.model.AnnotationMirrorImpl)this._env.getFactory().newAnnotationMirror((AnnotationBinding)jdtElementValue);
          returnVal = Proxy.newProxyInstance(expectedLeafType.getClassLoader(), 
              new Class[] { expectedLeafType }, annoMirror);
        } 
        Array.set(returnArray, i, returnVal);
      } else {
        Array.set(returnArray, i, null);
      } 
    } 
    return returnArray;
  }
  
  private Object convertJDTValueToReflectionType(Object jdtValue, TypeBinding actualType, Class<?> expectedType) {
    if (expectedType.isPrimitive() || String.class.equals(expectedType)) {
      if (jdtValue instanceof Constant) {
        if (boolean.class.equals(expectedType))
          return Boolean.valueOf(((Constant)jdtValue).booleanValue()); 
        if (byte.class.equals(expectedType))
          return Byte.valueOf(((Constant)jdtValue).byteValue()); 
        if (char.class.equals(expectedType))
          return Character.valueOf(((Constant)jdtValue).charValue()); 
        if (double.class.equals(expectedType))
          return Double.valueOf(((Constant)jdtValue).doubleValue()); 
        if (float.class.equals(expectedType))
          return Float.valueOf(((Constant)jdtValue).floatValue()); 
        if (int.class.equals(expectedType))
          return Integer.valueOf(((Constant)jdtValue).intValue()); 
        if (long.class.equals(expectedType))
          return Long.valueOf(((Constant)jdtValue).longValue()); 
        if (short.class.equals(expectedType))
          return Short.valueOf(((Constant)jdtValue).shortValue()); 
        if (String.class.equals(expectedType))
          return ((Constant)jdtValue).stringValue(); 
      } 
      return Factory.getMatchingDummyValue(expectedType);
    } 
    if (expectedType.isEnum()) {
      Object returnVal = null;
      if (actualType != null && actualType.isEnum() && jdtValue instanceof FieldBinding) {
        FieldBinding binding = (FieldBinding)jdtValue;
        try {
          Field returnedField = null;
          returnedField = expectedType.getField(new String(binding.name));
          if (returnedField != null)
            returnVal = returnedField.get(null); 
        } catch (NoSuchFieldException noSuchFieldException) {
        
        } catch (IllegalAccessException illegalAccessException) {}
      } 
      return (returnVal == null) ? Factory.getMatchingDummyValue(expectedType) : returnVal;
    } 
    if (expectedType.isAnnotation()) {
      if (actualType.isAnnotationType() && jdtValue instanceof AnnotationBinding) {
        org.eclipse.jdt.internal.compiler.apt.model.AnnotationMirrorImpl annoMirror = 
          (org.eclipse.jdt.internal.compiler.apt.model.AnnotationMirrorImpl)this._env.getFactory().newAnnotationMirror((AnnotationBinding)jdtValue);
        return Proxy.newProxyInstance(expectedType.getClassLoader(), 
            new Class[] { expectedType }, annoMirror);
      } 
      return null;
    } 
    return Factory.getMatchingDummyValue(expectedType);
  }
}
